(******************************************************************)
(*                                                                *)
(*                Svga HiColor Rendering Engine v1.4.2            *)
(*                                                                *)
(*  Copyright 1998 by Uvarov Dmitry                               *)
(*                                                                *)
(* Read MANUAL.TXT for more info                                  *)
(******************************************************************)
Unit SvgaLib;
interface
Uses DPMI,MathTMT,Mouse,crt;
{$I types.pas}
{$I def.pas}
CONST
  {library version}
  libVERSION = '1.4.2';

  LFBmode        = $80;

  CurrentMode:word=0;

  SPECIAL        = $81FF;
  LFB320x200x8   : integer = 0;
  LFB320x200x15  : integer = 0;
  LFB320x200x16  : integer = 0;

  LFB640x480x8   : integer = 0;
  LFB640x480x15  : integer = 0;
  LFB640x480x16  : integer = 0;

  LFB800x600x8   : integer = 0;
  LFB800x600x15  : integer = 0;
  LFB800x600x16  : integer = 0;

  LFB1024x768x8  : integer = 0;
  LFB1024x768x15 : integer = 0;
  LFB1024x768x16 : integer = 0;

  LFB1280x1024x8 : integer = 0;
  LFB1280x1024x15: integer = 0;
  LFB1280x1024x16: integer = 0;

  {use it as InitLFBMode(LFBXXXxXXXxXX+DontClearMemory) to initiate}
  { graphics mode without clearing video memory }
  DontClearMemory= $8000;
  ForceLFBMode   = $4000;

  {constants for dac type}
  RGBdac         =1;
  BGRdac         =2;

 VAR VESAInfo: TVBEInfo;
     VESAModeInfo: TVBEModeInfo;
     ModesListPtr: ^TModesInfoList;  {pointer to short modes info structure}
     TotalVideoMemory:dword;  {total memory in bytes}
     VideoPage,SplitScreen:TPageType;

     --------------------------------------------------------------
     -- this flag modyfied only on unit initialization stage     --
     -- if it doesn't contain true, then svgalib not initialized --
     -- successfully and cannot be used for rendering on video   --
     -- memory surface                                           --
     --------------------------------------------------------------
     VESA_Detected:boolean;
     {number of detected video modes}
     Total_Modes:integer;
     {number of detected video modes with lfb support}
     Total_LFB_Modes:integer;

 CONST
     SplitScrnScanLine:word=$FFFF;
     SplitScrnActive:boolean=False;
     SplitScrnHeight :word=0;
     GetMaxX         :word=0;
     GetMaxY         :word=0;
     GetMaxColor     :word=0;
     MaxRed          :word=0;
     MaxGreen        :word=0;
     MaxBlue         :word=0;
     ScrnSizeX       :word=0;
     ScrnSizeY       :word=0;
     PhysScrnSizeX   :word=0;
     PhysScrnSizeY   :word=0;
     WindowPosX      :word=0;  {this values assigned when SetStartAddr is called}
     WindowPosY      :word=0;  {they has physical coordinates of current view window}
     LFB_seg:dword=0;
     LFB_StartSeg:dword=0;   {if Split screen active, value of start segment}
                             {will be different, than LFB_seg. LFB_seg will}
                             {point to Split screen segment}
     BytesPerPixel   :integer=0;

     DACType         :integer=0;  {type of video card DAC, i.e. RGB or BGR}
     ClipTop         :fixed=-(32000 shl 16);
     ClipBot         :fixed=32000 shl 16;
     ClipLeft        :fixed=-(32000 shl 16);
     ClipRight       :fixed=32000 shl 16;
{$IFDEF Z_CLIPPING}
     {set Z clipping range to it's extrems}
     MinZClip        :fixed=-(32000 shl 16);
     MaxZClip        :fixed=32000 shl 16;
{$ENDIF}
 {error codes}
 CONST
  lfberr_OK      =0;
  lfberr_NoMemory=1;
  lfberr_LFBnotSupported=2;
  lfberr_NotVESA20compatible=3;
  lfberr_VESAnotDetected=4;
  lfberr_ModeNotSupported=5;
  lfberr_ErrorSettingMode=6;
  lfberr_NoModesFound   =7;
{$IFDEF DEBUG}
  VbeErrorMsg:array[0..7] of string[30]=('ok',
                                         'not enought video memory',
                                         'LFB mode not supported',
                                         'VESA 2.0 not detected',
                                         'VESA not detected',
                                         'mode not supported',
                                         'mode initialization error',
                                         'LFB modes not found');
{$ENDIF}

{returns false if mode not supported}
Function  GetModeInfo(Mode:word; var InfoBlock:TVbeModeInfo):boolean;
{this procedure initiate LFB mode}
Function  InitLFBMode(mode:word):byte; {returns error code}
{initiate LFB mode by given resolution and number if bits per pixel}
{returns error code}
Function  SetLFBMode(xres,yres,bpp:integer):byte;
Procedure CloseLFB;
Function  InitSplitScreen(rows:word):byte;
Procedure CloseSplitScreen;
Procedure AdjustSplitScreen(row:word);
Procedure ShowSplitScreen;
Procedure HideSplitScreen;
Procedure SetStartAddr(x,y:word);
Function  SetLogicalSize(x,y:word):word; {returns actual scan line lenght set}
Function  GetLogicalWidth:word;  {return current scan line width}
Function  GetRGBColor(r,g,b:byte):word;
{this procedures extract rgb components from color}
Procedure ExtractRGBColor(col:word; var dst:TRGBcolor);
Procedure ClearScreen(Where:TPageType;color:word);

Procedure PutPix8 (x,y:word; color:word; Where:TPageType);
Procedure PutPix16(x,y:word; color:word; Where:TPageType);
Function  GetPix8 (x,y:word; Where:TPageType):word;
Function  GetPix16(x,y:word; Where:TPageType):word;
Procedure Line8(x1,y1,x2,y2,color:word;where:TPageType);
Procedure Line16(x1,y1,x2,y2,color:word;where:TPageType);
Procedure Bar8(x1,y1,x2,y2,color:word;Where:TPageType);
Procedure Bar16(x1,y1,x2,y2,color:word;Where:TPageType);

var PutPix:procedure(x,y:word; color:word; where:TPageType);
    GetPix:function (x,y:word; where:TPageType):word;
    Line  :procedure(x1,y1,x2,y2,color:word;where:TPageType);
    Bar   :procedure(x1,y1,x2,y2,color:word;Where:TPageType);
{SVGA engine procedures and functions}
Procedure LoadTransparencyTable(Table:TTranspTablePtr);
--
-- calling to folloing routines must ne implemented in following way:
-- DrawPoly(@MyPoly, NumVertexes, MyColor, VideoPage);
--
Procedure DrawFlatPoly8 (poly:TPolygonPtr; Num:byte; color:word; Where:TPageType);
Procedure DrawFlatPoly16(poly:TPolygonPtr; Num:byte; color:word; Where:TPageType);

Procedure DrawTransparentFlatPoly8 (poly:TPolygonPtr; Num:byte; color:word; Where:TPageType);
Procedure DrawTransparentFlatPoly16(poly:TPolygonPtr; Num:byte; color:word; Where:TPageType);

Procedure DrawGoraudPoly8 (poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);
Procedure DrawGoraudPoly16(poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);

Procedure DrawTransparentGoraudPoly8 (poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);

Procedure RenderTexture8(poly:TPolygonPtr; txt:TTexture; Num:byte; Where:TPageType);
Procedure RenderTexture16(poly:TPolygonPtr; txt:TTexture; Num:byte; Where:TPageType);
Procedure RenderLambertTexture8(poly:TPolygonPtr; txt:TTexture; Num:byte; color:word; Where:TPageType);
Procedure RenderGoraudTexture8(poly:TPolygonPtr; txt:TTexture; Num:byte; color:word; Where:TPageType);

Procedure RenderArbitraryTexture8(poly:TPolygonPtr; txt:TTexture; Num:byte; Where:TPageType);

Function ClipPoly(Src,Dst:TPolygonPtr; Num:byte):byte;
Procedure SetClipWindow(x1,y1,x2,y2:integer);
{$IFDEF Z_CLIPPING}
Procedure SetZClippingRange(MinZ,MaxZ:integer);
{$ENDIF}
var DrawFlatPoly: procedure(poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);
    DrawTransparentFlatPoly: procedure(poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);
    DrawGoraudPoly: procedure(poly:TPolygonPtr; Num:byte; Color:word; Where:TPageType);
    RenderTexture: procedure(poly:TPolygonPtr; txt:TTexture; Num:byte; Where:TPageType);
    RenderLambertTexture: procedure(poly:TPolygonPtr; txt:TTexture; Num:byte; color:word; Where:TPageType);
    RenderGoraudTexture: procedure(poly:TPolygonPtr; txt:TTexture; Num:byte; color:word; Where:TPageType);
{mouse part procedures}
Const
   {if this flag set, then if cursor will come up to split screen, it will be}
   { drawn on it }
   DrawOverSplitScreen:boolean=True;

Procedure DrawMousePointer(x,y:integer);
Procedure RestoreMouseBK(x,y,WinX:integer);
Procedure SaveMouseBK(x,y,WinX:integer);
Procedure SetMouseWindow(x1,y1,x2,y2:integer);

Procedure LoadMousePointer(var p:TMousePointer);
Procedure FreeMousePointer(var p:TMousePointer);
Procedure SetMousePointer(p:TMousePointer);

Procedure ShowMousePointer;
Procedure HideMousePointer;
Procedure SetMouseWritePage(Where:TPageType);

Implementation
var regs:TRmRegs;
    _ColorAndMask:word;  {used to perform tranparency calculation in 15&16-bit modes}

{$I src\SvgaEng.pas -- graphics polygon engine}
{$I src\SvgaL8.pas  -- graphics procedures for 8-bit modes}
{$I src\SvgaL16.pas -- graphics procedures for 16&15-bit modes}
{$I src\SvgMouse.pas-- mouse procedures}

Function GetModeInfo(Mode:word; var InfoBlock:TVbeModeInfo):boolean;
 begin
    {this must be done for some VESA implementations}
    FillChar(pointer(buf_32)^,SizeOf(TVbeInfo),$00000000);

    ClearRmRegs(Regs);
    Regs.ES:=buf_16;  {use internal 8k buffer}
    Regs.AX:=$4F01;
    Regs.CX:=mode and $3FFF;  {clear all flags}
    RealModeInt($10,Regs);
    { if mode supported? }
    if (Regs.AX<>$004F) then
       GetModeInfo:=false
    else begin
       GetModeInfo:=true;
       Move(Pointer(buf_32)^,InfoBlock,SizeOf(TVbeModeInfo));
    end;
 end;

{***************************************************************************}
{this procedure called on unit initialization stage and not avaible for user}
function InitSvgaLib:integer;
  Type TList=array[1..100] of word;
  var
      a:dword;
      pword:^word;
      ModeInfo:TVbeModeInfo;
      mode:word;
      ModesList:^TList;
  begin
   Total_Modes:=0;
   Total_LFB_Modes:=0;
   result:=0;
   {force vbe to give VESA20 extended info}
   VESAInfo.VBESignature[1]:='V';
   VESAInfo.VBESignature[2]:='B';
   VESAInfo.VBESignature[3]:='E';
   VESAInfo.VBESignature[4]:='2';

   {this must be done for some VESA implementations}
   FillChar(pointer(buf_32)^,SizeOf(TVbeInfo),$00000000);
   {copy first 4 bytes that contain VBE2 string into temporary block}
   move(VESAInfo,pointer(buf_32)^,4);

   {check vesa compatibility}
   {simulate real mode interrupt}
   ClearRmRegs(Regs);
   Regs.ax:=$4F00;
   Regs.es:=buf_16;
   RealModeInt($10,Regs);
   move(pointer(buf_32)^,VESAinfo,SizeOf(TVbeInfo));
   if (Regs.ax<>$004F) or ((VESAinfo.VBESignature[1]<>'V') OR
                           (VESAinfo.VBESignature[2]<>'E') OR
                           (VESAinfo.VBESignature[3]<>'S') OR
                           (VESAinfo.VBESignature[4]<>'A')) then begin
      result:=lfberr_VesaNotDetected;
      exit;
   end;
   if VESAInfo.VBEVersion<$200 then
      result:=lfberr_NotVESA20compatible;
   {continue initiating svgalib and reading modes list}

   TotalVideoMemory:=VESAinfo.TotalMemory*65536;

{$IFDEF DEBUG}
   writeln('VESA 2.0 detected');
   write('reading available modes list...');
{$ENDIF}


   {now convert all realmode pointers to linear addresses}
   a:=(VESAinfo.OemStringPtr);
   VESAInfo.OemStringPtr:=((a AND $0FFFF)+(((a shr 16) AND $0FFFF) shl 4));
   a:=(VESAInfo.VideoModePtr);
   VESAInfo.VideoModePtr:=((a AND $0FFFF)+(((a shr 16) AND $0FFFF) shl 4));
{note that next fields will contain no information if video card not VESA 2.0 compatible}
   a:=(VESAInfo.OEMVendorNamePtr);
   VESAInfo.OEMVendorNamePtr:=((a AND $0FFFF)+(((a shr 16) AND $0FFFF) shl 4));
   a:=(VESAInfo.OEMProductNamePtr);
   VESAInfo.OEMProductNamePtr:=((a AND $0FFFF)+(((a shr 16) AND $0FFFF) shl 4));
   a:=(VESAInfo.OEMProductRevPtr);
   VESAInfo.OEMProductRevPtr:=((a AND $0FFFF)+(((a shr 16) AND $0FFFF) shl 4));

   {get total number of all modes}
     pword:=pointer(VESAInfo.VideoModePtr);
     if dword(pword)=0 then begin
         writeln('fatal error: modes list not found...program halted');
         halt(1);
     end;

     {get memory for temorary modes list}
     GetMem(ModesList,200*2);
     a:=0;
     {if here some modes?}
     if pword^<>$0FFFF then
       repeat
           a+:=1;
           ModesList^[a]:=pword^;
           pword:=pword+2;
       until pword^=$0FFFF;
     Total_Modes:=a;
     {allocate memory for modes structure}
     if Total_Modes<>0 then GetMem(ModesListPtr,SizeOf(TShortModeInfo)*Total_Modes);

   {now get short info about all modes and set default constants}
   pword:=pointer(VESAInfo.VideoModePtr);
   for a:=1 to Total_Modes do begin
       mode:=ModesList^[a];
       if GetModeInfo(mode,ModeInfo) then begin
          {collect short info about this mode}
          ModesListPtr^[a].ModeNumber:=mode;
          ModesListPtr^[a].XRes:=ModeInfo.XResolution;
          ModesListPtr^[a].YRes:=ModeInfo.YResolution;
          ModesListPtr^[a].BitsPerPixel:=ModeInfo.BitsPerPixel;
          ModesListPtr^[a].ModeAttributes:=ModeInfo.ModeAttributes;
          {this mode support lfb?}
          if (ModeInfo.ModeAttributes AND LFBMode)<>0 then Total_LFB_Modes+:=1;
          {now setup default constants}
          mode:=mode OR ForceLFBmode;
          with ModeInfo do
            case XResolution of
              320: if YResolution=200 then
                       if BitsPerPixel=8 then LFB320x200x8:=mode else
                       if BitsPerPixel=15 then LFB320x200x15:=mode else
                       if BitsPerPixel=16 then LFB320x200x16:=mode;
              640: if YResolution=480 then
                       if BitsPerPixel=8 then LFB640x480x8:=mode else
                       if BitsPerPixel=15 then LFB640x480x15:=mode else
                       if BitsPerPixel=16 then LFB640x480x16:=mode;
              800: if YResolution=600 then
                       if BitsPerPixel=8 then LFB800x600x8:=mode else
                       if BitsPerPixel=15 then LFB800x600x15:=mode else
                       if BitsPerPixel=16 then LFB800x600x16:=mode;
             1024: if YResolution=768 then
                       if BitsPerPixel=8 then LFB1024x768x8:=mode else
                       if BitsPerPixel=15 then LFB1024x768x15:=mode else
                       if BitsPerPixel=16 then LFB1024x768x16:=mode;
             1280: if YResolution=1024 then
                       if BitsPerPixel=8 then LFB1280x1024x8:=mode else
                       if BitsPerPixel=15 then LFB1280x1024x15:=mode else
                       if BitsPerPixel=16 then LFB1280x1024x16:=mode;
            end;

       end else FillChar(ModesListPtr^[a],SizeOf(TShortModeInfo),0);

       pword:=pword+2;
   end;

   {release allocated memory}
   FreeMem(ModesList,200*2);

   
   if Total_Modes=0 then result:=lfberr_NoModesFound;
{$IFDEF DEBUG}
   writeln('done...Detected ',Total_Modes,' modes');
{$ENDIF}

  end;
{***************************************************************************}
Function  GetRGBColor(r,g,b:byte):word;assembler;
  asm
    mov al,r
    mov cl,VesaModeInfo.RedFieldPosition
    and ax,MaxRed
    mov bl,g
    shl ax,cl

    and bx,MaxGreen
    mov cl,VesaModeInfo.GreenFieldPosition
    shl bx,cl
    mov dl,b
    mov cl,VesaModeInfo.BlueFieldPosition

    and dx,MaxBlue
    or  ax,bx
    shl dx,cl
    or  ax,dx
  end;
{***************************************************************************}
Procedure ExtractRGBColor(col:word; var dst:TRGBcolor);assembler;
  asm
    mov ax,col
    mov cl,VesaModeInfo.RedFieldPosition
    mov dx,ax
    shr ax,cl
    mov edi,dst
    and ax,MaxRed
    mov cl,VesaModeInfo.GreenFieldPosition
    mov bx,dx
    mov [edi+2],al    {store red component}
    shr bx,cl
    mov cl,VesaModeInfo.BlueFieldPosition
    and bx,MaxGreen
    shr dx,cl
    mov [edi+1],bl    {store green component}
    and dx,MaxBlue
    mov [edi],dl     {store blue component}
  end;
{***************************************************************************}
Procedure ClearScreen(Where:TPageType;color:word);
begin
  asm
    sub eax,eax
    sub edx,edx
    mov edi,dword ptr Where
    -- calculate screen size
    mov dx,word ptr Where[6]
    mov ax,word ptr Where[4]
    mul edx
    mov ecx,eax

    cmp BytesPerPixel,1  -- if it a 8-bit mode
    je  @@fill8bit
    shr ecx,1     -- calc screen size in dword's
    mov ax,color
    jmp @@fill
  @@fill8bit:
    mov al,byte ptr color
    mov ah,al
    shr ecx,2
  @@fill:
    mov bx,ax     -- fill color pattern in eax
    shl eax,16
    mov ax,bx

    rep stosd
  end;
end;
{***************************************************************************}
Function  SetLFBMode(xres,yres,bpp:integer):byte;
  var i:integer;
  begin
    result:=0;
    for i:=1 to Total_Modes do
       if (ModesListPtr^[i].xres=xres) AND (ModesListPtr^[i].yres=yres) AND
                  (ModesListPtr^[i].BitsPerPixel=bpp) then begin
              result:=InitLFBMode(ModesListPtr^[i].ModeNumber);
              exit;
       end;
    result:=lfberr_ModeNotSupported;
  end;
{***************************************************************************}
Function InitLFBMode(mode:word):byte;
begin
{$IFDEF DEBUG}
     writeln('trying to initialize mode ',mode,'...');
{$ENDIF}
  if (not VESA_Detected) then begin
{$IFDEF DEBUG}
     writeln('VESA 2.0 not detected...could not initiate video mode');
{$ENDIF}
     InitLFBMode:=lfberr_ErrorSettingMode;
     exit;
  end;

  if (not GetModeInfo(mode,VESAModeInfo)) or (mode=0) then begin
  (* mode not supported *)
{$IFDEF DEBUG}
     writeln('selected mode not supported');
{$ENDIF}
    InitLFBMode:=lfberr_ModeNotSupported;
    exit;
  end;

  if VesaModeInfo.BitsPerPixel=8 then BytesPerPixel:=1 else BytesPerPixel:=2;
  if TotalVideoMemory<(VesaModeInfo.XResolution*VesaModeInfo.YResolution*BytesPerPixel) then begin
{$IFDEF DEBUG}
    writeln('not enought video memory to initiate mode');
{$ENDIF}
    InitLFBMode:=lfberr_noMemory;
    exit;
  end;
  { Map physical address, returned in VESAModeInfo.PhysBasePtr }
  { by function 4F01h. Set limit on TotalVideoMemory+1 Mb for any case }
{$ifdef DEBUG}
  write('locating linar framebuffer at ',VESAModeInfo.PhysBasePtr,'...');
{$endif}
  LFB_seg:=MapPhysicalToLinear(VESAModeInfo.PhysBasePtr,TotalVideoMemory+1024*1024);

  if LFB_seg=0 then begin
     initLFBmode:=lfberr_LFBnotSupported;
{$ifdef DEBUG}
     writeln('error');
{$endif}
     exit;
  end;
{$ifdef DEBUG}
  writeln('done');
  write('initiating mode ',VESAModeInfo.XResolution,
                       'x',VESAModeInfo.YResolution,
                       'x',VESAModeInfo.BitsPerPixel,'...');
{$endif}

  (* Set VESA VBE video mode *)
  ClearRmRegs(Regs);
  Regs.AX:=$4F02;
  Regs.BX:=mode or ForceLFBmode;  {set lfb flag}
  RealModeInt($10,Regs);
  (* Is mode installed corectly ? *)
  if (Regs.AX<>$004F) then begin
    initLFBmode:=lfberr_ErrorSettingMode;
{$ifdef DEBUG}
    writeln('error');
    writeln('closing linear framebuffer...done');
{$endif}
    FreePhysicalMap(LFB_seg);
    exit;
  end;
{$ifdef DEBUG}
  writeln('done');
{$endif}

  (* set variables *)
  LFB_StartSeg:=LFB_seg;
  CurrentMode:=mode AND $3FFF; {clear all flags}
  ScrnSizeX:=VesaModeInfo.XResolution;
  ScrnSizeY:=VesaModeInfo.YResolution;
  GetMaxX:=ScrnSizeX-1;
  GetMaxY:=ScrnSizeY-1;
  PhysScrnSizeX:=ScrnSizeX;
  PhysScrnSizeY:=ScrnSizeY;
  if VesaModeInfo.BlueFieldPosition=0 then DACtype:=BGRdac else DACtype:=RGBdac;

  with VideoPage do begin
       seg:=LFB_StartSeg; SizeX:=ScrnSizeX; SizeY:=ScrnSizeY;
  end;
  SplitScreen:=VideoPage;

  if BytesPerPixel=1 then begin
     PutPix:=PutPix8;
     GetPix:=GetPix8;
     Bar:=Bar8;
     Line:=Line8;
     DrawFlatPoly:=DrawFlatPoly8;
     DrawGoraudPoly:=DrawGoraudPoly8;
     DrawTransparentFlatPoly:=DrawTransparentFlatPoly8;
     RenderTexture:=RenderTexture8;
     RenderLambertTexture:=RenderLambertTexture8;
     RenderGoraudTexture:=RenderGoraudTexture8;

     GetMaxColor:=255;
  end
  else begin
     PutPix:=PutPix16;
     GetPix:=GetPix16;
     Bar:=Bar16;
     Line:=Line16;
     DrawFlatPoly:=DrawFlatPoly16;
     DrawGoraudPoly:=DrawGoraudPoly16;
     DrawTransparentFlatPoly:=DrawTransparentFlatPoly16;
     RenderTexture:=RenderTexture16;
     RenderLambertTexture:=RenderLambertTexture16;
     RenderGoraudTexture:=RenderGoraudTexture16;

     GetMaxColor:=1 shl VesaModeInfo.BitsPerPixel-1;
     MaxRed  :=1 shl VesaModeInfo.RedMaskSize  -1;
     MaxGreen:=1 shl VesaModeInfo.GreenMaskSize-1;
     MaxBlue :=1 shl VesaModeInfo.BlueMaskSize -1;
     {this used to perform tranparency calculation in 16-bit mode}
     _ColorAndMask:=(MaxBlue shr 1) shl VesaModeInfo.BlueFieldPosition+
                    (MaxGreen shr 1) shl VesaModeInfo.GreenFieldPosition+
                    (MaxRed shr 1) shl VesaModeInfo.RedFieldPosition;
  end;

  InitLFBMode:=lfberr_OK;
end;

Procedure CloseLFB;
 begin
  if CurrentMode<>0 then begin
  (* Set standard VGA text mode (03h BIOS) *)
     ClearRmRegs(Regs);
     Regs.AX:=$03;
     RealModeInt($10,Regs);

     (* Free physical mapping *)
     FreePhysicalMap(LFB_seg);
     CurrentMode:=0;
  end;
 end;

Procedure SetStartAddr(x,y:word);
  var offs:word;
  begin
    Regs.BX:=0;
    Regs.AX:=$04F07;
    Regs.DX:=y+SplitScrnHeight;
    Regs.CX:=x;
    RealModeInt($10,Regs);
    WindowPosX:=x;
    WindowPosY:=y;
    {if mouse active then update mouse pointer}
    {this only take effect then DrawOverSplitScreen=true and split screen}
    {visible and mouse located over it}
--    if GetButtonCount<>0 then UpdateMousePointer;
  end;

Function InitSplitScreen(rows:word):byte;
 begin
   {is there enougth video memory?}
  with VESAModeInfo do begin
   if ((rows+ScrnSizeY)*BytesPerPixel*ScrnSizeX)>TotalVideoMemory then begin
      InitSplitScreen:=lfberr_NoMemory;
      exit;
   end;
   LFB_StartSeg:=LFB_seg+BytesPerPixel*ScrnSizeX*rows;
   SplitScrnHeight:=rows;
   SetStartAddr(0,0);

   VideoPage.seg:=LFB_StartSeg;
   SplitScreen.seg:=LFB_seg;
   SplitScreen.SizeY:=rows;

   InitSplitScreen:=lfberr_OK;
  end;
 end;

Procedure CloseSplitScreen;
 {remove split screen and restore variables}
 begin
   LFB_StartSeg:=LFB_seg;
   SplitScrnHeight:=0;
   SplitScrnActive:=false;
   SetStartAddr(0,0);
   VideoPage.seg:=LFB_seg;
   SplitScreen.SizeY:=0;
 end;

Procedure AdjustSplitScreen(row:word);
    var a:byte;
    begin
       {wait for horizontal retrace}

       while Port[$3DA] AND 1 <> 0 do;
       while Port[$3Da] AND 1 = 0 do;

       Port[$3D4]:=$18;
       Port[$3D5]:=Lo(row);  {write first 8 bits}
       Port[$3D4]:=$07;
       Port[$3D5]:=(Port[$3D5] and $EF) or ((row shr 4) and $10);  {write 9th bit}
       Port[$3D4]:=$09;
       Port[$3D5]:=(Port[$3D5] and $BF) or ((row shr 3) and $40);  {write 10t bit}

(*
  Port[$3D4]:=$18; Port[$3D5]:=Lo(row);  { Bit 0-7, Line compare }
  Port[$3D4]:=$07; A:=Port[$3D5];      { Bit 8, Overflow }
  if row<$100 then A:=A and $EF else A:=A or $10;
  Port[$3D5]:=A;
  Port[$3D4]:=$09; A:=Port[$3D5];      { Bit 9, Maximum scan line }
  if row<$200 then A:=A and $BF else A:=A or $40;
  Port[$3D5]:=A;
*)
       if Row<ScrnSizeY then
          SplitScrnActive:=True
       else SplitScrnActive:=False;
       SplitScrnScanLine:=row;
    end;
Procedure ShowSplitScreen;
  begin
    AdjustSplitScreen(GetMaxY-SplitScrnHeight);
  end;
Procedure HideSplitScreen;
  begin
    AdjustSplitScreen($FFFF);
  end;

Function  SetLogicalSize(x,y:word):word;
  begin
    {set logical screen size}
    Regs.AX:=$4F06;
    Regs.BX:=0;
    Regs.CX:=x;
    RealModeInt($10,Regs); { dx:=maximum number of scan lines }
    SetLogicalSize:=Regs.CX;

    ScrnSizeX:=Regs.CX;
    ScrnSizeY:=y;
    GetMaxX:=ScrnSizeX-1;
    GetMaxY:=ScrnSizeY-1;

    {update video page start address}
    LFB_StartSeg:=LFB_seg+BytesPerPixel*ScrnSizeX*SplitScrnHeight;
    VideoPage.seg:=LFB_StartSeg;
    VideoPage.SizeX:=ScrnSizeX;
    VideoPage.SizeY:=ScrnSizeY;

    SplitScreen.SizeX:=ScrnSizeX;
    SetStartAddr(0,0);
  end;

Function GetLogicalWidth:word;
  begin
    Regs.AX:=$4F06;
    Regs.BX:=1;
    RealModeInt($10,regs);
    GetLogicalWidth:=Regs.CX;
  end;

var OldExitProc:pointer;
Procedure _CloseSvgaLib;
  begin
     if CurrentMode<>0 then CloseLFB;
{$ifdef DEBUG}
     writeln;
     writeln('------------------------------------------------------------------------');
     if ErrorAddr<>Nil then writeln('Extreme exit from Svga engine...');
     writeln('closing Svga HiColor rendering engine...done');
{$endif}
     if Total_Modes<>0 then FreeMem(ModesListPtr,SizeOf(TShortModeInfo)*Total_Modes);
     ExitProc:=OldExitProc;
  end;

--Procedure _RunTimeErrorExit(exitCode,r_ebp,r_eip:dword);
-- begin
--   writeln('RunError #',exitCode,' at ',r_eip,'... Extreme exit from SvgaLib');
--   halt(exitCode);
-- end;

Begin
  OldExitProc:=ExitProc;
  ExitProc:=@_CloseSvgaLib;

-- {$System}
--  System.%runerr_call_stack_proc_addr := @_RunTimeErrorExit;

{$IFDEF SHOW_LOGO}
  writeln('Svga HiColor rendering engine v'+libVERSION+' by Uvarov Dmitry (C)1998-1999');
  writeln('------------------------------------------------------------------------');
{$ENDIF}
{$IFDEF DEBUG}
  writeln('initializing Svga engine...');
  write('checking VESA 2.0 compatibility...');
{$ENDIF}
  DECLARE var er:integer; begin
     er:=InitSvgaLib;
     if er<>0 then begin
{$IFDEF DEBUG}
        writeln('error: ',VbeErrorMsg[er]);
        writeln('initialization failed');
        writeln('------------------------------------------------------------------------');
{$ENDIF}
        VESA_Detected:=false;
     end else begin
{$IFDEF DEBUG}
        writeln('initialization completed');
        writeln('------------------------------------------------------------------------');
{$ENDIF}
        VESA_Detected:=true;
     end;
  end;

End.